@@ -1,6 +1,6 @@
 # This file is a part of pyctr.
 #
-# Copyright (c) 2017-2020 Ian Burgwin
+# Copyright (c) 2017-2021 Ian Burgwin
 # This file is licensed under The MIT License (MIT).
 # You can find the full license text in LICENSE in the root of this project.
 
@@ -47,6 +47,8 @@ DEV_COMMON_KEY_0 = bytes.fromhex('55A3F872BDC80C555A654381139E153B')
 
 MIN_TICKET_SIZE = 0x2AC
 
+OTP_MAGIC = b'\x0f\xb0\xad\xde'
+
 
 class CryptoError(PyCTRError):
     """Generic exception for cryptography operations."""
@@ -213,6 +215,15 @@ def _requires_bootrom(method):
     return wrapper
 
 
+def _requires_otp(method):
+    @wraps(method)
+    def wrapper(self, *args, **kwargs):
+        if not self.b9_keys_set:
+            raise KeyslotMissingError('an OTP dump is required, see setup_keys_from_otp')
+        return method(self, *args, **kwargs)
+    return wrapper
+
+
 # used from http://www.falatic.com/index.php/108/python-and-bitwise-rotation
 # converted to def because pycodestyle complained to me
 def rol(val: int, r_bits: int, max_bits: int) -> int:
@@ -249,12 +260,15 @@ class CryptoEngine:
     :param setup_b9_keys: Whether or not to automatically load keys from boot9.
     """
 
-    __slots__ = ['key_x', 'key_y', 'key_normal', 'dev', 'b9_keys_set', 'b9_path', '_b9_extdata_otp',
-                 '_b9_extdata_keygen', '_otp_device_id', '_otp_key', '_otp_iv', '_id0']
+    __slots__ = ['key_x', 'key_y', 'key_normal', 'dev', 'b9_keys_set', 'b9_path', 'otp_keys_set', '_otp_enc',
+                 '_otp_dec', '_b9_extdata_otp', '_b9_extdata_keygen', '_otp_device_id', '_otp_key', '_otp_iv', '_id0']
 
     b9_keys_set: bool
     """Keys have been set from the ARM9 BootROM."""
 
+    otp_keys_set: bool
+    """Keys have been set from a dumped OTP region."""
+
     b9_path: 'Optional[str]'
     """
     Path that the ARM9 BootROM was loaded from. Set when :meth:`setup_keys_from_boot9_file` is called, which is
@@ -276,6 +290,8 @@ class CryptoEngine:
         self.b9_keys_set = False
         self.b9_path = None
 
+        self.otp_keys_set = False
+
         self._b9_extdata_otp: Optional[bytes] = None
         self._b9_extdata_keygen: Optional[bytes] = None
 
@@ -283,6 +299,9 @@ class CryptoEngine:
         self._otp_key: Optional[bytes] = None
         self._otp_iv: Optional[bytes] = None
 
+        self._otp_enc: Optional[bytes] = None
+        self._otp_dec: Optional[bytes] = None
+
         self._id0: Optional[bytes] = None
 
         for keyslot, keys in base_key_x.items():
@@ -301,11 +320,6 @@ class CryptoEngine:
     def b9_extdata_keygen(self) -> bytes:
         return self._b9_extdata_keygen
 
-    @property
-    @_requires_bootrom
-    def otp_device_id(self) -> int:
-        return self._otp_device_id
-
     @property
     @_requires_bootrom
     def otp_key(self) -> bytes:
@@ -316,6 +330,21 @@ class CryptoEngine:
     def otp_iv(self) -> bytes:
         return self._otp_iv
 
+    @property
+    @_requires_otp
+    def otp_device_id(self) -> int:
+        return self._otp_device_id
+
+    @property
+    @_requires_otp
+    def otp_dec(self) -> bytes:
+        return self._otp_dec
+
+    @property
+    @_requires_otp
+    def otp_enc(self) -> bytes:
+        return self._otp_enc
+
     @property
     def id0(self) -> bytes:
         if not self._id0:
@@ -391,29 +420,46 @@ class CryptoEngine:
 
         return CMAC.new(key, ciphermod=AES)
 
-    def create_ctr_io(self, keyslot: Keyslot, fh: 'BinaryIO', ctr: int):
+    def create_ctr_io(self, keyslot: Keyslot, fh: 'BinaryIO', ctr: int, closefd: bool = False):
         """
         Create an AES-CTR read-write file object with the given keyslot.
 
         :param keyslot: :class:`Keyslot` to use.
         :param fh: File-like object to wrap.
         :param ctr: Counter to start with.
+        :param closefd Close underlying file object when closed.
         :return: A file-like object that does decryption and encryption on the fly.
         :rtype: CTRFileIO
         """
-        return CTRFileIO(fh, self, keyslot, ctr)
+        if keyslot < 0x04:
+            return TWLCTRFileIO(file=fh,
+                                crypto=self,
+                                keyslot=keyslot,
+                                counter=ctr,
+                                closefd=True)
+        else:
+            return CTRFileIO(file=fh,
+                             crypto=self,
+                             keyslot=keyslot,
+                             counter=ctr,
+                             closefd=closefd)
 
-    def create_cbc_io(self, keyslot: Keyslot, fh: 'BinaryIO', iv: bytes):
+    def create_cbc_io(self, keyslot: Keyslot, fh: 'BinaryIO', iv: bytes, closefd: bool = False):
         """
         Create an AES-CBC read-only file object with the given keyslot.
 
         :param keyslot: :class:`Keyslot` to use.
         :param fh: File-like object to wrap.
         :param iv: Initialization vector.
+        :param closefd Close underlying file object when closed.
         :return: A file-like object that does decryption on the fly.
         :rtype: CBCFileIO
         """
-        return CBCFileIO(fh, self, keyslot, iv)
+        return CBCFileIO(file=fh,
+                         crypto=self,
+                         keyslot=keyslot,
+                         iv=iv,
+                         closefd=closefd)
 
     @staticmethod
     def sd_path_to_iv(path: str) -> int:
@@ -648,30 +694,33 @@ class CryptoEngine:
             raise OTPLengthError(otp_len)
 
         cipher_otp = AES.new(self.otp_key, AES.MODE_CBC, self.otp_iv)
-        if otp[0:4] == b'\x0f\xb0\xad\xde':
+        if otp[0:4] == OTP_MAGIC:
             # decrypted otp
-            otp_enc: bytes = cipher_otp.encrypt(otp)
-            otp_dec = otp
+            self._otp_enc: bytes = cipher_otp.encrypt(otp)
+            self._otp_dec = otp
         else:
             # encrypted otp
-            otp_enc = otp
-            otp_dec: bytes = cipher_otp.decrypt(otp)
+            self._otp_enc = otp
+            self._otp_dec: bytes = cipher_otp.decrypt(otp)
+        
+        if self._otp_dec[0:4] != OTP_MAGIC:
+            raise CorruptOTPError('OTP magic not found, corrupt or not an OTP')
 
-        self._otp_device_id = int.from_bytes(otp_dec[4:8], 'little')
+        self._otp_device_id = int.from_bytes(self.otp_dec[4:8], 'little')
 
-        otp_hash: bytes = otp_dec[0xE0:0x100]
-        otp_hash_digest: bytes = sha256(otp_dec[0:0xE0]).digest()
+        otp_hash: bytes = self.otp_dec[0xE0:0x100]
+        otp_hash_digest: bytes = sha256(self.otp_dec[0:0xE0]).digest()
         if otp_hash_digest != otp_hash:
             raise CorruptOTPError(f'expected: {otp_hash.hex()}; result: {otp_hash_digest.hex()}')
 
-        otp_keysect_hash: bytes = sha256(otp_enc[0:0x90]).digest()
+        otp_keysect_hash: bytes = sha256(self.otp_enc[0:0x90]).digest()
 
         self.set_keyslot('x', Keyslot.New3DSKeySector, otp_keysect_hash[0:0x10])
         self.set_keyslot('y', Keyslot.New3DSKeySector, otp_keysect_hash[0x10:0x20])
 
         # most otp code from https://github.com/Stary2001/3ds_tools/blob/master/three_ds/aesengine.py
 
-        twl_cid_lo, twl_cid_hi = readle(otp_dec[0x08:0xC]), readle(otp_dec[0xC:0x10])
+        twl_cid_lo, twl_cid_hi = readle(self.otp_dec[0x08:0xC]), readle(self.otp_dec[0xC:0x10])
         twl_cid_lo ^= 0xB358A6AF
         twl_cid_lo |= 0x80000000
         twl_cid_hi ^= 0x08C267B7
@@ -679,7 +728,7 @@ class CryptoEngine:
         twl_cid_hi = twl_cid_hi.to_bytes(4, 'little')
         self.set_keyslot('x', 0x03, twl_cid_lo + b'NINTENDO' + twl_cid_hi)
 
-        console_key_xy: bytes = sha256(otp_dec[0x90:0xAC] + self.b9_extdata_otp).digest()
+        console_key_xy: bytes = sha256(self.otp_dec[0x90:0xAC] + self.b9_extdata_otp).digest()
         self.set_keyslot('x', Keyslot.Boot9Internal, console_key_xy[0:0x10])
         self.set_keyslot('y', Keyslot.Boot9Internal, console_key_xy[0x10:0x20])
 
@@ -733,6 +782,8 @@ class CryptoEngine:
             self.set_keyslot('x', i, d[off:off + 16])
             off += 16
 
+        self.otp_keys_set = True
+
     @_requires_bootrom
     def setup_keys_from_otp_file(self, path: 'Union[PathLike, str, bytes]'):
         """Set up console-unique keys from an OTP file. Encrypted and decrypted are supported."""
@@ -767,9 +818,12 @@ class _CryptoFileBase(RawIOBase):
 
     closed = False
     _reader: 'BinaryIO'
+    _closefd: bool
 
     def close(self):
         self.closed = True
+        if self._closefd:
+            self._reader.close()
 
     __del__ = close
 
@@ -801,15 +855,21 @@ class _CryptoFileBase(RawIOBase):
 class CTRFileIO(_CryptoFileBase):
     """Provides transparent read-write AES-CTR encryption as a file-like object."""
 
-    def __init__(self, file: 'BinaryIO', crypto: 'CryptoEngine', keyslot: Keyslot, counter: int):
+    def __init__(self, file: 'BinaryIO', crypto: 'CryptoEngine', keyslot: Keyslot, counter: int,
+                 closefd: bool = False):
         self._reader = file
         self._crypto = crypto
         self._keyslot = keyslot
         self._counter = counter
+        self._closefd = closefd
         self._lock = Lock()
 
     def __repr__(self):
-        return f'{type(self).__name__}(file={self._reader!r}, keyslot={self._keyslot}, counter={self._counter!r})'
+        return (f'{type(self).__name__}(file={self._reader!r}, keyslot={self._keyslot}, counter={self._counter!r}, '
+                f'closefd={self._closefd!r})')
+
+    def __hash__(self):
+        return hash((self._reader, self._keyslot, self._counter))
 
     @_raise_if_file_closed
     def read(self, size: int = -1) -> bytes:
@@ -841,18 +901,54 @@ class CTRFileIO(_CryptoFileBase):
         return self._reader.truncate(size)
 
 
+class TWLCTRFileIO(CTRFileIO):
+    """Provides transparent read-write TWL AES-CTR encryption as a file-like object."""
+
+    # TWL AES operations need data added at the end too
+    # because each 0x10-byte block is flipped before and after it is de/encrypted
+
+    @_raise_if_file_closed
+    def read(self, size: int = -1) -> bytes:
+        with self._lock:
+            cur_offset = self.tell()
+            data = self._reader.read(size)
+            padding_before = cur_offset % 0x10
+            padding_after = (-(padding_before + len(data)) % 0x10)
+            counter = self._counter + (cur_offset >> 4)
+            cipher = self._crypto.create_ctr_cipher(self._keyslot, counter)
+            data = (b'\0' * padding_before) + data + (b'\0' * padding_after)
+            return cipher.decrypt(data)[padding_before:len(data) - padding_after]
+
+    @_raise_if_file_closed
+    def write(self, data: bytes) -> int:
+        with self._lock:
+            cur_offset = self.tell()
+            padding_before = cur_offset % 0x10
+            padding_after = (-(padding_before + len(data)) % 0x10)
+            counter = self._counter + (cur_offset >> 4)
+            cipher = self._crypto.create_ctr_cipher(self._keyslot, counter)
+            data = (b'\0' * padding_before) + data + (b'\0' * padding_after)
+            data = cipher.encrypt(data)
+            return self._reader.write(data[padding_before:len(data) - padding_after])
+
+
 class CBCFileIO(_CryptoFileBase):
     """Provides transparent read-only AES-CBC encryption as a file-like object."""
 
-    def __init__(self, file: 'BinaryIO', crypto: 'CryptoEngine', keyslot: Keyslot, iv: bytes):
+    def __init__(self, file: 'BinaryIO', crypto: 'CryptoEngine', keyslot: Keyslot, iv: bytes, closefd: bool = False):
         self._reader = file
         self._crypto = crypto
         self._keyslot = keyslot
         self._iv = iv
+        self._closefd = closefd
         self._lock = Lock()
 
     def __repr__(self):
-        return f'{type(self).__name__}(file={self._reader!r}, keyslot={self._keyslot}, iv={self._iv!r})'
+        return (f'{type(self).__name__}(file={self._reader!r}, keyslot={self._keyslot}, iv={self._iv!r}, '
+                f'closefd={self._closefd!r})')
+
+    def __hash__(self):
+        return hash((self._reader, self._keyslot, self._iv))
 
     @_raise_if_file_closed
     def read(self, size: int = -1):
