@@ -1,6 +1,6 @@
 # This file is a part of pyctr.
 #
-# Copyright (c) 2017-2020 Ian Burgwin
+# Copyright (c) 2017-2021 Ian Burgwin
 # This file is licensed under The MIT License (MIT).
 # You can find the full license text in LICENSE in the root of this project.
 
@@ -213,7 +213,7 @@ class ExeFSReader(TypeReaderBase):
 
         self.entries = {}
 
-        header = fp.read(EXEFS_HEADER_SIZE)
+        header = self._file.read(EXEFS_HEADER_SIZE)
 
         # ExeFS entries can fit up to 10 names. hashes are stored in reverse order
         #   (e.g. the first entry would have the hash at the very end - 0x1E0)
