@@ -1,12 +1,12 @@
 # This file is a part of pyctr.
 #
-# Copyright (c) 2017-2020 Ian Burgwin
+# Copyright (c) 2017-2021 Ian Burgwin
 # This file is licensed under The MIT License (MIT).
 # You can find the full license text in LICENSE in the root of this project.
 
 """Module for interacting with encrypted SD card contents under the "Nintendo 3DS" directory."""
 
-from os import PathLike
+from os import PathLike, fsdecode
 from pathlib import Path
 from typing import TYPE_CHECKING
 
@@ -65,7 +65,7 @@ class SDFilesystem:
         elif sd_key_file:
             self._crypto.setup_sd_key_from_file(sd_key_file)
 
-        self._base_path = Path(path)
+        self._base_path = Path(fsdecode(path))
 
         try:
             self._id0_path = self._base_path / self._crypto.id0.hex()
